# This file is part of pex_config.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["ChoiceField"]

from .callStack import getStackFrame
from .config import Field, FieldTypeVar, _typeStr


class ChoiceField(Field[FieldTypeVar]):
    """A configuration field (`~lsst.pex.config.Field` subclass) that allows a
    user to select from a predefined set of values.

    Use ``ChoiceField`` when a configuration can only take one of a predefined
    set of values. Each choice must be of the same type.

    Parameters
    ----------
    doc : `str`
        Documentation string that describes the configuration field.
    dtype : class
        The type of the field's choices. For example, `str` or `int`.
    allowed : `dict`
        The allowed values. Keys are the allowed choices (of ``dtype``-type).
        Values are descriptions (`str`-type) of each choice.
    default : ``dtype``-type, optional
        The default value, which is of type ``dtype`` and one of the allowed
        choices.
    optional : `bool`, optional
        If `True`, this configuration field is *optional*. Default is `True`.
    deprecated : None or `str`, optional
        A description of why this Field is deprecated, including removal date.
        If not None, the string is appended to the docstring for this Field.

    See also
    --------
    ConfigChoiceField
    ConfigDictField
    ConfigField
    ConfigurableField
    DictField
    Field
    ListField
    RangeField
    RegistryField
    """

    def __init__(self, doc, dtype=None, allowed=None, default=None, optional=True, deprecated=None):
        if dtype is None:
            raise ValueError("supplied dtype must not be None")
        if allowed is None:
            raise ValueError("supplied allowed mapping must not be None")
        self.allowed = dict(allowed)
        if optional and None not in self.allowed:
            self.allowed[None] = "Field is optional"

        if len(self.allowed) == 0:
            raise ValueError("ChoiceFields must allow at least one choice")

        Field.__init__(
            self, doc=doc, dtype=dtype, default=default, check=None, optional=optional, deprecated=deprecated
        )

        self.__doc__ += "\n\nAllowed values:\n\n"
        for choice, choiceDoc in self.allowed.items():
            if choice is not None and not isinstance(choice, dtype):
                raise ValueError(
                    "ChoiceField's allowed choice %s is of incorrect type %s. Expected %s"
                    % (choice, _typeStr(choice), _typeStr(dtype))
                )
            self.__doc__ += "%s\n  %s\n" % ("``{0!r}``".format(str(choice)), choiceDoc)

        self.source = getStackFrame()

    def _validateValue(self, value):
        Field._validateValue(self, value)
        if value not in self.allowed:
            msg = "Value {} is not allowed.\n\tAllowed values: [{}]".format(
                value, ", ".join(str(key) for key in self.allowed)
            )
            raise ValueError(msg)



# This file is part of pex_config.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ("Color", "format")

import os
import re
import sys


class Color:
    """A controller that determines whether strings should be colored.

    Parameters
    ----------
    text : `str`
        Text content to print to a terminal.
    category : `str`
        Semantic category of the ``text``. See `categories` for possible
        values.

    Raises
    ------
    RuntimeError
        Raised when the ``category`` is not a key of ``Color.categories``.

    Notes
    -----
    The usual usage is ``Color(string, category)`` which returns a string that
    may be printed; categories are given by the keys of `Color.categories`.

    `Color.colorize` may be used to set or retrieve whether the user wants
    color. It always returns `False` when `sys.stdout` is not attached to a
    terminal.
    """

    categories = dict(
        NAME="blue",
        VALUE="yellow",
        FILE="green",
        TEXT="red",
        FUNCTION_NAME="blue",
    )
    """Mapping of semantic labels to color names (`dict`).

    Notes
    -----
    The default categories are:

    - ``'NAME'``
    - ``'VALUE'``
    - ``'FILE'``
    - ``'TEXT'``
    - ``'FUNCTION_NAME'``
    """

    colors = {
        "black": 0,
        "red": 1,
        "green": 2,
        "yellow": 3,
        "blue": 4,
        "magenta": 5,
        "cyan": 6,
        "white": 7,
    }
    """Mapping of color names to terminal color codes (`dict`).
    """

    _colorize = True

    def __init__(self, text, category):
        try:
            color = Color.categories[category]
        except KeyError:
            raise RuntimeError("Unknown category: %s" % category)

        self.rawText = str(text)
        x = color.lower().split(";")
        self.color, bold = x.pop(0), False
        if x:
            props = x.pop(0)
            if props in ("bold",):
                bold = True

        try:
            self._code = "%s" % (30 + Color.colors[self.color])
        except KeyError:
            raise RuntimeError("Unknown colour: %s" % self.color)

        if bold:
            self._code += ";1"

    @staticmethod
    def colorize(val=None):
        """Get or set whether the string should be colorized.

        Parameters
        ----------
        val : `bool` or `dict`, optional
            The value is usually a bool, but it may be a dict which is used
            to modify Color.categories

        Returns
        -------
        shouldColorize : `bool`
            If `True`, the string should be colorized. A string **will not** be
            colorized if standard output or standard error are not attached to
            a terminal or if the ``val`` argument was `False`.

            Only strings written to a terminal are colorized.
        """

        if val is not None:
            Color._colorize = val

            if isinstance(val, dict):
                unknown = []
                for k in val:
                    if k in Color.categories:
                        if val[k] in Color.colors:
                            Color.categories[k] = val[k]
                        else:
                            print("Unknown colour %s for category %s" % (val[k], k), file=sys.stderr)
                    else:
                        unknown.append(k)

                if unknown:
                    print("Unknown colourizing category: %s" % " ".join(unknown), file=sys.stderr)

        return Color._colorize if sys.stdout.isatty() else False

    def __str__(self):
        if not self.colorize():
            return self.rawText

        base = "\033["

        prefix = base + self._code + "m"
        suffix = base + "m"

        return prefix + self.rawText + suffix


def _colorize(text, category):
    text = Color(text, category)
    return str(text)


def format(config, name=None, writeSourceLine=True, prefix="", verbose=False):
    """Format the history record for a configuration, or a specific
    configuration field.

    Parameters
    ----------
    config : `lsst.pex.config.Config`
        A configuration instance.
    name : `str`, optional
        The name of a configuration field to specifically format the history
        for. Otherwise the history of all configuration fields is printed.
    writeSourceLine : `bool`, optional
        If `True`, prefix each printout line with the code filename and line
        number where the configuration event occurred. Default is `True`.
    prefix : `str`, optional
        A prefix for to add to each printout line. This prefix occurs first,
        even before any source line. The default is an empty string.
    verbose : `bool`, optional
        Default is `False`.
    """

    if name is None:
        for i, name in enumerate(config.history.keys()):
            if i > 0:
                print()
            print(format(config, name))

    outputs = []
    for value, stack, label in config.history.get(name, []):
        output = []
        for frame in stack:
            if frame.function in (
                "__new__",
                "__set__",
                "__setattr__",
                "execfile",
                "wrapper",
            ) or os.path.split(frame.filename)[1] in ("argparse.py", "argumentParser.py"):
                if not verbose:
                    continue

            line = []
            if writeSourceLine:
                line.append(
                    [
                        "%s" % ("%s:%d" % (frame.filename, frame.lineno)),
                        "FILE",
                    ]
                )

            line.append(
                [
                    frame.content,
                    "TEXT",
                ]
            )
            if False:
                line.append(
                    [
                        frame.function,
                        "FUNCTION_NAME",
                    ]
                )

            output.append(line)

        outputs.append([value, output])

    if outputs:
        # Find the maximum widths of the value and file:lineNo fields.
        if writeSourceLine:
            sourceLengths = []
            for value, output in outputs:
                sourceLengths.append(max([len(x[0][0]) for x in output]))
            sourceLength = max(sourceLengths)

        valueLength = len(prefix) + max([len(str(value)) for value, output in outputs])

    # Generate the config history content.
    msg = []
    fullname = "%s.%s" % (config._name, name) if config._name is not None else name
    msg.append(_colorize(re.sub(r"^root\.", "", fullname), "NAME"))
    for value, output in outputs:
        line = prefix + _colorize("%-*s" % (valueLength, value), "VALUE") + " "
        for i, vt in enumerate(output):
            if writeSourceLine:
                vt[0][0] = "%-*s" % (sourceLength, vt[0][0])

            output[i] = " ".join([_colorize(v, t) for v, t in vt])

        line += ("\n%*s" % (valueLength + 1, "")).join(output)
        msg.append(line)

    return "\n".join(msg)



# This file is part of pex_config.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test that we can load a YAML config of a Config class that we have
not previously imported.
"""

import unittest

try:
    import yaml
except ImportError:
    yaml = None

# We have to import the core package to allow the standardized
# yaml constructor to be registered for all subclasses
import lsst.pex.config  # noqa: F401

# The serialized YAML form was created by doing:
# from ticketDM26008helper.unloaded import Unloaded
# c = Unloaded()
# c.c = "World"
# print(yaml.dump(c))

serialized = """
!<lsst.pex.config.Config> |
  import ticketDM26008helper.unloaded
  assert type(config)==ticketDM26008helper.unloaded.Unloaded, 'config is of type %s.%s instead of ticketDM26008helper.unloaded.Unloaded' % (type(config).__module__, type(config).__name__)
  # integer test
  config.i=None

  # float test
  config.f=3.0

  # boolean test
  config.b=False

  # choice test
  config.c='World'

  # Range test
  config.r=3.0

  # list test
  config.ll=[1, 2, 3]

  # dict test
  config.d={'key': 'value'}

  # nan test
  config.n=float('nan')
"""  # noqa: E501


class UnloadedYaml(unittest.TestCase):
    @unittest.skipIf(yaml is None, "pyyaml not available")
    def testLoadUnloaded(self):
        loaded = yaml.safe_load(serialized)
        self.assertEqual(loaded.c, "World")


if __name__ == "__main__":
    unittest.main()



# This file is part of pex_config.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle
import unittest

import lsst.pex.config as pexConfig


class Config1(pexConfig.Config):
    d1 = pexConfig.DictField("d1", keytype=str, itemtype=int, default={"hi": 4}, itemCheck=lambda x: x > 0)
    d2 = pexConfig.DictField("d2", keytype=str, itemtype=str, default=None)
    d3 = pexConfig.DictField("d3", keytype=float, itemtype=float, optional=True, itemCheck=lambda x: x > 0)
    d4 = pexConfig.DictField("d4", keytype=str, itemtype=None, default={})


class DictFieldTest(unittest.TestCase):
    def testConstructor(self):
        try:

            class BadKeytype(pexConfig.Config):
                d = pexConfig.DictField("...", keytype=list, itemtype=int)

        except Exception:
            pass
        else:
            raise SyntaxError("Unsupported keyptype DictFields should not be allowed")

        try:

            class BadItemtype(pexConfig.Config):
                d = pexConfig.DictField("...", keytype=int, itemtype=dict)

        except Exception:
            pass
        else:
            raise SyntaxError("Unsupported itemtype DictFields should not be allowed")

        try:

            class BadItemCheck(pexConfig.Config):
                d = pexConfig.DictField("...", keytype=int, itemtype=int, itemCheck=4)

        except Exception:
            pass
        else:
            raise SyntaxError("Non-callable itemCheck DictFields should not be allowed")

        try:

            class BadDictCheck(pexConfig.Config):
                d = pexConfig.DictField("...", keytype=int, itemtype=int, dictCheck=4)

        except Exception:
            pass
        else:
            raise SyntaxError("Non-callable dictCheck DictFields should not be allowed")

    def testFieldTypeAnnotationRuntime(self):
        # test parsing type annotation for runtime keytype, itemtype
        testField = pexConfig.DictField[str, int](doc="")
        self.assertEqual(testField.keytype, str)
        self.assertEqual(testField.itemtype, int)

        # verify that forward references work correctly
        testField = pexConfig.DictField["float", "int"](doc="")
        self.assertEqual(testField.keytype, float)
        self.assertEqual(testField.itemtype, int)

        # verify that Field rejects single types
        with self.assertRaises(ValueError):
            pexConfig.DictField[int](doc="")  # type: ignore

        # verify that Field raises in conflict with keytype, itemtype
        with self.assertRaises(ValueError):
            pexConfig.DictField[str, int](doc="", keytype=int)

        with self.assertRaises(ValueError):
            pexConfig.DictField[str, int](doc="", itemtype=str)

        # verify that Field does not raise if dtype agrees
        testField = pexConfig.DictField[int, str](doc="", keytype=int, itemtype=str)
        self.assertEqual(testField.keytype, int)
        self.assertEqual(testField.itemtype, str)

    def testAssignment(self):
        c = Config1()
        self.assertRaises(pexConfig.FieldValidationError, setattr, c, "d1", {3: 3})
        self.assertRaises(pexConfig.FieldValidationError, setattr, c, "d1", {"a": 0})
        self.assertRaises(pexConfig.FieldValidationError, setattr, c, "d1", [1.2, 3, 4])
        c.d1 = None
        c.d1 = {"a": 1, "b": 2}
        self.assertRaises(pexConfig.FieldValidationError, setattr, c, "d3", {"hi": True})
        c.d3 = {4: 5}
        self.assertEqual(c.d3, {4.0: 5.0})
        d = {"a": None, "b": 4, "c": "foo"}
        c.d4 = d
        self.assertEqual(c.d4, d)
        c.d4["a"] = 12
        c.d4["b"] = "three"
        c.d4["c"] = None
        self.assertEqual(c.d4["a"], 12)
        self.assertEqual(c.d4["b"], "three")
        self.assertIsNone(c.d4["c"])
        self.assertRaises(pexConfig.FieldValidationError, setattr, c, "d4", {"hi": [1, 2, 3]})

    def testValidate(self):
        c = Config1()
        self.assertRaises(pexConfig.FieldValidationError, Config1.validate, c)

        c.d2 = {"a": "b"}
        c.validate()

    def testInPlaceModification(self):
        c = Config1()
        self.assertRaises(pexConfig.FieldValidationError, c.d1.__setitem__, 2, 0)
        self.assertRaises(pexConfig.FieldValidationError, c.d1.__setitem__, "hi", 0)
        c.d1["hi"] = 10
        self.assertEqual(c.d1, {"hi": 10})

        c.d3 = {}
        c.d3[4] = 5
        self.assertEqual(c.d3, {4.0: 5.0})

    def testNoArbitraryAttributes(self):
        c = Config1()
        self.assertRaises(pexConfig.FieldValidationError, setattr, c.d1, "should", "fail")

    def testEquality(self):
        """Test DictField.__eq__

        We create two dicts, with the keys explicitly added in a different
        order and test their equality.
        """
        keys1 = ["A", "B", "C"]
        keys2 = ["X", "Y", "Z", "a", "b", "c", "d", "e"]

        c1 = Config1()
        c1.d4 = {k: "" for k in keys1}
        for k in keys2:
            c1.d4[k] = ""

        c2 = Config1()
        for k in keys2 + keys1:
            c2.d4[k] = ""

        self.assertTrue(pexConfig.compareConfigs("test", c1, c2))

    def testNoPickle(self):
        """Test that pickle support is disabled for the proxy container."""
        c = Config1()
        with self.assertRaises(pexConfig.UnexpectedProxyUsageError):
            pickle.dumps(c.d4)


if __name__ == "__main__":
    unittest.main()



# This file is part of pex_config.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
import tempfile
import unittest

import lsst.pex.config as pexConf


class PexTestConfig(pexConf.Config):
    list1 = pexConf.ListField(dtype=int, default=[1, 2], doc="list1")
    f1 = pexConf.Field(dtype=float, doc="f1")
    f2 = pexConf.Field(dtype=float, doc="f2")


class EqualityTest(unittest.TestCase):
    def test(self):
        c1 = PexTestConfig()
        c2 = PexTestConfig()
        self.assertEqual(c1, c2)
        c1.list1 = [1, 2, 3, 4, 5]
        self.assertNotEqual(c1, c2)
        c2.list1 = c1.list1
        self.assertEqual(c1, c2)


class LoadSpecialTest(unittest.TestCase):
    def test(self):
        c1 = PexTestConfig()
        c2 = PexTestConfig()
        c1.list1 = None
        c1.f1 = float("nan")
        c2.f2 = float("inf")
        with tempfile.NamedTemporaryFile(mode="w") as f:
            c1.saveToStream(f.file)
            f.file.close()
            c2.load(f.name)
        self.assertEqual(c1.list1, c2.list1)
        self.assertEqual(c1.f2, c2.f2)
        self.assertTrue(math.isnan(c2.f1))


if __name__ == "__main__":
    unittest.main()
